#!/usr/bin/python

import os
import sys
from errors import CommandError
from hunk_selector import ShelveHunkSelector, UnshelveHunkSelector
from diffstat import DiffStat
from patchsource import PatchSource, FilePatchSource

BASE_DIR = '.shelf'

class Shelf(object):
    MESSAGE_PREFIX = "# Shelved patch: "

    def __init__(self, base, name='default'):
        self.name = name
        self.base = base
        shelf_base = os.path.join(self.base, BASE_DIR)
        self.shelf_dir = os.path.join(shelf_base, name)

        for dir in [shelf_base, self.shelf_dir]:
            if not os.path.isdir(dir):
                os.mkdir(dir)

    def log(self, msg):
        print >>sys.stderr, msg

    def list(self):
        self.log("Patches on shelf '%s':" % self.name)
        for patch in self.__list():
            msg = self.get_patch_message(self.__path(patch))
            if msg is None:
                msg = "No message saved with patch."
            self.log(' %.2d: %s' % (patch, msg))

    def __path(self, idx):
        return os.path.join(self.shelf_dir, '%.2d' % idx)

    def next_patch(self):
        index = 0
        while True:
            name = self.__path(index)
            if not os.path.exists(name):
                return name
            index += 1

    def __list(self):
        patches = os.listdir(self.shelf_dir)
        patches = [int(f) for f in patches]
        patches.sort()
        return patches

    def last_patch(self):
        patches = self.__list()

        if len(patches) == 0:
            return None

        return self.__path(patches[-1])

    def get_patch_message(self, patch_path):
        patch = open(patch_path, 'r').read()

        if not patch.startswith(self.MESSAGE_PREFIX):
            return None
        return patch[len(self.MESSAGE_PREFIX):patch.index('\n')]

    def __show_status(self, source):
        if source.can_live_update():
            diff_stat = str(DiffStat(source.readlines()))
            if len(diff_stat) > 0:
                self.log('Diff status is now:\n' + diff_stat)
            else:
                self.log('No changes left in working tree.')

    def unshelve(self, patch_source, pick_hunks=False):
        patch_name = self.last_patch()

        if patch_name is None:
            raise CommandError("No patch found on shelf %s" % self.name)

        hunks = FilePatchSource(patch_name).readhunks()
        if pick_hunks:
            to_unshelve, to_remain = UnshelveHunkSelector(hunks).select()
        else:
            to_unshelve = hunks
            to_remain = []

        if len(to_unshelve) == 0:
            raise CommandError('Nothing to unshelve')

        message = self.get_patch_message(patch_name)
        if message is None:
            message = ""
        self.log('Reapplying shelved patches "%s"' % message)

        pipe = os.popen('patch -d %s -s -p0' % self.base, 'w')
        for hunk in to_unshelve:
            pipe.write(str(hunk))
        pipe.flush()

        if pipe.close() is not None:
            raise CommandError("Failed running patch!")

        if len(to_remain) == 0:
            os.remove(patch_name)
        else:
            f = open(patch_name, 'w')
            for hunk in to_remain:
                f.write(str(hunk))
            f.close()

        self.__show_status(patch_source)

    def shelve(self, patch_source, pick_hunks=False, message=None):
        from datetime import datetime

        hunks = patch_source.readhunks()

        if pick_hunks:
            to_shelve = ShelveHunkSelector(hunks).select()[0]
        else:
            to_shelve = hunks

        if len(to_shelve) == 0:
            raise CommandError('Nothing to shelve')

        if message is None:
            timestamp = datetime.now().strftime('%Y-%m-%d %H:%M:%S')
            message = "Changes shelved on %s" % timestamp

        patch_name = self.next_patch()
        self.log('Shelving to %s/%s: "%s"' % \
                (self.name, os.path.basename(patch_name), message))

        patch = open(patch_name, 'a')

        assert '\n' not in message
        patch.write("%s%s\n" % (self.MESSAGE_PREFIX, message))

        for hunk in to_shelve:
            patch.write(str(hunk))

        patch.flush()
        os.fsync(patch.fileno())
        patch.close()

        pipe = os.popen('patch -d %s -sR -p0' % self.base, 'w')
        for hunk in to_shelve:
            pipe.write(str(hunk))
        pipe.flush()

        if pipe.close() is not None:
            raise CommandError("Failed running patch!")

        self.__show_status(patch_source)
